from ._version import __version__  # noqa

from .citation import (
    CiteClass,
    cite_function,
    cite_module,
    cite_object,
    find_in_citations,
    get_all_citations,
    get_used_citations,
    print_all_citations,
    print_used_citations,
    reset_used_citations,
)

from .import_utils import get_all_imports

__all__ = [
    "CiteClass",
    "cite_function",
    "cite_module",
    "cite_object",
    "find_in_citations",
    "get_all_citations",
    "get_all_imports",
    "get_used_citations",
    "print_all_citations",
    "print_used_citations",
    "reset_used_citations",
]



"""Two sample benchmarks to compute runtime and memory usage.

For more information on writing benchmarks:
https://asv.readthedocs.io/en/stable/writing_benchmarks.html."""

import citation_compass as cc


@cc.cite_function("fake")
def fake_function():
    """A fake function to demonstrate the use of the citation_compass package."""
    return 1


@cc.cite_function("fake", track_used=False)
def fake_function2():
    """A fake function to demonstrate the use of the citation_compass package."""
    return 1


class FakeClass(cc.CiteClass):
    """A fake class for benchmarking."""

    def __init__(self):
        pass


def time_create_function():
    """Time the use of a wrapper with a label."""

    @cc.cite_function("example")
    def test_function():
        return 1


def time_call_function():
    """Time the cost of calling a wrapped function."""
    _ = fake_function()


def time_call_function_untracked():
    """Time the cost of calling an untracked function."""
    _ = fake_function2()


def time_instantiate_class():
    """Time the cost of instantiating a class with a wrapped method."""
    _ = FakeClass()


def time_cite_module():
    """Time the cost of citing a module."""
    cc.cite_module("citation_compass")



"""Fake module for testing.

Citation: CitationCompass, 2025.
"""

from citation_compass import cite_module, cite_function, CiteClass

cite_module(__name__)


def fake_uncited_function():
    """This is a fake function for testing purposes."""
    return 0


class FakeClass:
    """A fake class for testing."""

    def __init__(self):
        pass

    @cite_function
    def fake_method(self):
        """A fake class method for testing."""
        return 0


class FakeCitedClass(CiteClass):
    """A 2nd fake class for testing."""

    def __init__(self):
        pass

    def fake_method(self):
        """A fake (uncited) class method for testing."""
        return 1


# Test that we propogate the citation to subclasses.
class InheritedFakeClass(FakeCitedClass):
    """A 3rd fake class for testing."""

    def __init__(self):
        pass



"""A helper module to collect citations from a software package."""

from functools import wraps
import inspect
import logging
import sys

from citation_compass.docstring_utils import (
    extract_citation,
    extract_urls,
)

CITATION_REGISTRY_ALL = {}
CITATION_REGISTRY_USED = set()


def _get_full_name(obj):
    """Return the maximally qualified name of a thing.

    Parameters
    ----------
    obj : object
        The obj to get the name of.

    Returns
    -------
    str
        The fully qualified name of the thing.
    """
    # Try to determine the name of the "thing".
    if hasattr(obj, "__qualname__"):
        base_name = obj.__qualname__
    elif hasattr(obj, "__name__"):
        base_name = obj.__name__
    elif hasattr(obj, "__class__"):
        # If this is an object, use the class's name.
        base_name = obj.__class__.__qualname__
    else:
        raise ValueError(f"Could not determine the name of {obj}")

    # Get the string for the module (if we can find it).
    module = inspect.getmodule(obj)
    full_name = base_name if module is None else f"{module.__name__}.{base_name}"
    return full_name


class CitationEntry:
    """A (data)class to store information about a citation.

    Attributes
    ----------
    key : str
        The name of the module, function, or other aspect where the citation is needed.
    citation : str, optional
        The citation string.
    label : str, optional
        The (optional) user-defined label for the citation.
    urls : list of str
        A list of URLs extracted from the citation string.
    """

    def __init__(self, key, citation=None, label=None):
        self.key = key
        self.citation = citation
        self.label = label

        if citation is None:
            if label is not None and len(label) > 0:
                self.citation = label
            else:
                self.citation = "No citation provided."

        self.urls = extract_urls(self.citation)

    def __hash__(self):
        return hash(self.key)

    def __str__(self):
        return f"{self.key}: {self.citation}"

    def __repr__(self):
        return f"{self.key}:\n{self.citation}"

    @classmethod
    def from_object(cls, obj, label=None):
        """Create a CitationEntry from any object (including a function or method).

        Parameters
        ----------
        obj : object
            The object from which to create the citation.
        label : str, optional
            The (optional) user-defined label for the citation.

        Returns
        -------
        CitationEntry
            The citation entry.
        """
        # Try to parse a citation from the object's docstring (if there is one).
        if hasattr(obj, "__doc__"):
            docstring = obj.__doc__
        elif hasattr(obj, "__class__") and hasattr(obj.__class__, "__doc__"):
            docstring = obj.__class__.__doc__
        else:
            docstring = ""
        citation_text = extract_citation(docstring)
        if citation_text is None:
            citation_text = docstring

        full_name = _get_full_name(obj)

        return cls(
            key=full_name,
            citation=citation_text,
            label=label,
        )


def cite_module(name, citation=None):
    """Add a citation to a entire module.

    Parameters
    ----------
    name : str
        The name of the module.
    citation : str, optional
        The citation string. If None the code automatically tries to extract the citation text
        from the module's docstring.
    """
    if citation is None and name in sys.modules:
        module = sys.modules[name]
        if hasattr(module, "__doc__"):
            citation = extract_citation(module.__doc__)
            if citation is None or len(citation) == 0:
                citation = module.__doc__

    CITATION_REGISTRY_ALL[name] = CitationEntry(name, citation)
    CITATION_REGISTRY_USED.add(name)


class CiteClass:
    """A super class for adding a citation to a class."""

    def __init__(self):
        pass

    def __init_subclass__(cls):
        # Add the class's full name
        full_name = _get_full_name(cls)
        cls._citation_compass_name = full_name

        # Save the citation as ALL when it is first defined.
        if full_name not in CITATION_REGISTRY_ALL:
            CITATION_REGISTRY_ALL[full_name] = CitationEntry.from_object(cls)
        else:
            logging.warning(f"Duplicated citation tag for class: {full_name}")

        # Wrap the constructor so the class is marked used when
        # the first object is instantiated.
        original_init = cls.__init__

        @wraps(original_init)
        def init_wrapper(*args, **kwargs):
            # Save the citation as USED when it is first called.
            if cls._citation_compass_name not in CITATION_REGISTRY_USED:
                CITATION_REGISTRY_USED.add(cls._citation_compass_name)
            return original_init(*args, **kwargs)

        cls.__init__ = init_wrapper


def cite_function(label=None, track_used=True):
    """A function wrapper for adding a citation to a function or
    class method.

    Parameters
    ----------
    label : str, optional
        The (optional) user-defined label for the citation.
    track_used : bool
        If True, the function will be marked as used when it is called.
        This adds a small amount of overhead to each function call.
        Default: True.

    Returns
    -------
    function
        The wrapped function or method.
    """
    # If the label is callable, there were no parentheses on the
    # dectorator and it passed in the function instead. So use None
    # as the label.
    use_label = label if not callable(label) else None

    def decorator(func):
        full_name = _get_full_name(func)

        # Save the citation as ALL when it is first defined.
        if full_name not in CITATION_REGISTRY_ALL:
            citation = CitationEntry.from_object(func, label=use_label)
            CITATION_REGISTRY_ALL[full_name] = citation
        else:
            logging.warning(f"Duplicated citation tag for function: {full_name}")

        # Wrap the function so it is marked as USED when it is called.
        if track_used:

            @wraps(func)
            def fun_wrapper(*args, **kwargs):
                # Save the citation as USED when it is first called.
                if func.__qualname__ not in CITATION_REGISTRY_USED:
                    CITATION_REGISTRY_USED.add(full_name)
                return func(*args, **kwargs)
        else:
            # We do not wrap the function, but just return the original function.
            fun_wrapper = func

            # We mark as used be default so the citation does not get dropped.
            CITATION_REGISTRY_USED.add(full_name)

        return fun_wrapper

    if callable(label):
        return decorator(label)
    return decorator


def cite_object(obj, label=None):
    """Add a citation for a specific object.

    Parameters
    ----------
    obj : object
        The object to add a citation for.
    label : str, optional
        The (optional) user-defined label for the citation.
    """
    full_name = _get_full_name(obj)
    if full_name not in CITATION_REGISTRY_ALL:
        CITATION_REGISTRY_ALL[full_name] = CitationEntry.from_object(obj, label=label)
    else:
        logging.warning(f"Duplicated citation tag for object: {full_name}")

    CITATION_REGISTRY_USED.add(full_name)


def get_all_citations():
    """Return a list of all citations in the software package.

    Returns
    -------
    citations : list of str
        A list of all citations in the software package.
    """
    citations = [str(entry) for entry in CITATION_REGISTRY_ALL.values()]
    return citations


def get_used_citations():
    """Return a list of all citations in the software package.

    Returns
    -------
    list of str
        A list of all citations in the software package.
    """
    citations = [str(CITATION_REGISTRY_ALL[func_name]) for func_name in CITATION_REGISTRY_USED]
    return citations


def find_in_citations(query, used_only=False):
    """Find a query string in the citation text. This is primarily used for
    testing, where a user might want to check if a citation is present.

    Parameters
    ----------
    query : str
        The query string to search for.
    used_only : bool, optional
        If True, only search in the used citations. If False, search in all citations.

    Returns
    -------
    matches : list of str
        A list of matching citation strings. This list is empty if no matches are found.
    """
    search_set = CITATION_REGISTRY_USED if used_only else CITATION_REGISTRY_ALL.keys()

    matches = []
    for name in search_set:
        entry = str(CITATION_REGISTRY_ALL[name])
        if query in entry:
            matches.append(entry)
    return matches


def reset_used_citations():
    """Reset the list of used citations."""
    CITATION_REGISTRY_USED.clear()


def print_all_citations():
    """Print all citations in the software package in a user-friendly way."""
    for name, citation in CITATION_REGISTRY_ALL.items():
        print(f"{name}:\n{citation.citation}\n")


def print_used_citations():
    """Print the used citations in the software package in a user-friendly way."""
    for name in CITATION_REGISTRY_USED:
        print(f"{name}:\n{CITATION_REGISTRY_ALL[name].citation}\n")



import fake_module

from citation_compass.import_utils import get_all_imports


def test_get_all_imports():
    """Check that the imports are registered."""
    imports = get_all_imports(skip_common=False)
    assert len(imports) > 0
    assert "sys" in imports
    assert "fake_module" in imports

    # We can filter out Python's base imports.
    imports = get_all_imports(skip_common=True)
    assert len(imports) > 0
    assert "sys" not in imports
    assert "fake_module" in imports

    # We can search for citation keywords in the module docstrings.
    old_len = len(imports)
    imports = get_all_imports(skip_common=True, use_keywords=True)
    assert len(imports) > 0
    assert len(imports) < old_len
    assert "fake_module" in imports

    # We call fake_uncited_function to make sure it is correctly imported.
    assert fake_module.fake_uncited_function() == 0



"""A helper module for searching docstrings for citations."""

_CITATION_HEADER_KEYWORDS = [
    "citation",
    "citations",
    "reference",
    "references",
]
_CITATION_OTHER_KEYWORDS = [
    "acknowledgement",
    "acknowledgements",
    "arxiv",
    "attribute",
    "attribution",
    "bibliography",
    "cite",
]

_CITATION_SECTION_HEADERS = set([f"{keyword}:" for keyword in _CITATION_HEADER_KEYWORDS])
_CITATION_ALL_KEYWORDS = _CITATION_HEADER_KEYWORDS + _CITATION_OTHER_KEYWORDS


def check_for_any_citation_keyword(string):
    """Checks a string for any of the keywords that indicate a citation,
    which can include the words in the middle of a sentence. As such, this approach
    is a heuristic and does not require the citation to be in a specific format. It
    is meant to be used to assess whether a module that is not tagged by this one
    may contain citations.

    Parameters
    ----------
    string : str
        The string to check.

    Returns
    -------
    bool
        Whether the docstring contains a keyword that indicates a citation.
    """
    if string is None or len(string) == 0:
        return False

    for line in [line.lower() for line in string.split("\n")]:
        for keyword in _CITATION_ALL_KEYWORDS:
            if keyword in line:
                return True
    return False


def extract_citation(docstring):
    """Extracts citation(s) from a docstring.

    This function assumes that the citation is in the formatted to
    match this package using the "keyword: information" structure at
    the start of a line or underlined keyword.

    For example, if the docstring contains either

    "Citation:
        Author, Title, year.

    Other information..."

    or

    "Citation
     --------
        Author, Title, year.

    Other information..."

    The extracted citation will be "Author, Title, year".

    Parameters
    ----------
    docstring : str
        The docstring to extract the citation from.

    Returns
    -------
    extracted_citations: str or None
        The extracted citation or None if no citation is found.
    """
    if docstring is None or len(docstring) == 0:
        return None

    all_lines = [line.strip() for line in docstring.split("\n")]

    # We search for a line the starts with one of the citation keywords.
    extracted_citations = ""
    current_citation = ""
    block_type = "None"
    for idx, line in enumerate(all_lines):
        if block_type == "None":
            # We are not currently in a citation block, so we check if this
            # line is the start of that block.
            if (
                len(line) > 1
                and line == "-" * len(line)
                and idx > 0
                and all_lines[idx - 1].lower() in _CITATION_HEADER_KEYWORDS
            ):
                # Check for an underlined section header. Note the section title will
                # be on the previous line.
                current_citation = ""
                block_type = "-"
            else:
                # Check if we are in a citation block of the form 'keyword: information'.
                for keyword in _CITATION_SECTION_HEADERS:
                    if line.lower().startswith(keyword):
                        block_type = ":"
                        current_citation = line[len(keyword) + 1 :]
                        break
        else:
            # We are already in a block, so check for the end of the block.
            if block_type == "-" and idx < len(all_lines) - 1:
                # Section type citation blocks end when we hit the next section.
                next_line = all_lines[idx + 1]

                if len(next_line) > 1 and next_line == "-" * len(next_line):
                    # Check by looking for the underline of the section header. If found,
                    # do not add the current line because it is the next section title.
                    block_type = "Done"
                elif len(line) == 0 and len(next_line) == 0:
                    # Sections can also end with two blank lines.
                    block_type = "Done"
            elif block_type == ":" and len(line) == 0:
                # End the 'keyword:' block when we hit a blank line.
                block_type = "Done"

            if block_type != "Done":
                # We are still in the block, so add the line to the current citation.
                current_citation += " " + line

        # We have finished the block, so add the citation to the extracted_citations.
        # We do this outside the if-else block to ensure we get single line citations.
        if block_type == "Done" or idx == len(all_lines) - 1:
            if len(current_citation) > 0:
                extracted_citations += "\n" + current_citation.strip()
                current_citation = ""

            # We are no longer in a citation block.
            block_type = "None"

    # Remove leading and trailing whitespace.
    extracted_citations = extracted_citations.strip()
    if len(extracted_citations) > 0:
        return extracted_citations
    return None


def extract_urls(string):
    """Extracts URLs from a string.

    Parameters
    ----------
    string : str
        The string to extract URLs from.

    Returns
    -------
    urls : list of str
        The extracted URLs.
    """
    if string is None or len(string) == 0:
        return []

    urls = []
    for word in string.split():
        if "http" in word:
            urls.append(word)
    return urls



"""A helper module for searching imports for citations."""

import sys

from citation_compass.docstring_utils import check_for_any_citation_keyword


def get_all_imports(skip_common=True, use_keywords=False):
    """Return a list of all imports in the software package.

    Parameters
    ----------
    skip_common : bool
        Whether to skip the common imports, such as the built-in modules.
    use_keywords : bool
        Uses a heuristic that checks if the import's docstring contains
        keywords that may indicate a citation.

    Returns
    -------
    imports : list of str
        A list of all imports in the software package.
    """
    imports = []
    for name, module in sys.modules.items():
        skip = False
        if name == "__main__":
            # Skip the main module
            skip = True
        elif hasattr(module, "__spec__") and module.__spec__ is not None:
            # Skip the built-in modules or ones from the python framework.
            origin = module.__spec__.origin
            if origin is None:
                skip = False
            elif origin == "built-in":
                skip = True
            elif origin == "frozen":
                skip = True
            elif "Python.framework" in origin:
                skip = True

        if not skip_common or not skip:
            if use_keywords and hasattr(module, "__doc__"):
                if check_for_any_citation_keyword(module.__doc__):
                    imports.append(name)
            else:
                imports.append(name)
    return imports



from citation_compass.docstring_utils import (
    _CITATION_ALL_KEYWORDS,
    check_for_any_citation_keyword,
    extract_citation,
    extract_urls,
)


def test_check_docstring_for_keyword():
    """Check that the function correctly identifies citation keywords."""
    assert check_for_any_citation_keyword("This is a docstring.") is False
    assert check_for_any_citation_keyword("This is a citation.") is True
    assert check_for_any_citation_keyword("This is a reference.") is True

    for keyword in _CITATION_ALL_KEYWORDS:
        assert check_for_any_citation_keyword(f"{keyword}: other stuff") is True

    # Check empty docstrings.
    assert check_for_any_citation_keyword("") is False
    assert check_for_any_citation_keyword(None) is False


def test_extract_citation_colon():
    """Test that we can extract a citation using the 'keyword:' format."""
    # Check an empty docstring.
    assert extract_citation("") is None
    assert extract_citation(None) is None

    # Start with single line docstrings.
    assert extract_citation("Citation: Author, Title, year.") == "Author, Title, year."
    assert extract_citation("Reference: Author, Title, year.") == "Author, Title, year."
    assert extract_citation("Info: Nothing to see here") is None

    # Test multi-line docstrings.
    docstring = """Top material:
    Stuff here.

    Citation:
        Author1, Author2, Title, year.

    Bottom material:
    More stuff."""
    assert extract_citation(docstring) == "Author1, Author2, Title, year."

    docstring = """Function description.

    Reference:
        Author1, Author2, Title, year.

    Parameters
    ----------
    Stuff here.

    Returns
    -------
    More stuff."""
    assert extract_citation(docstring) == "Author1, Author2, Title, year."

    docstring = """Function description.

    Parameters
    ----------
    Stuff here.

    Returns
    -------
    More stuff.

    Citation:
        Author1, Author2,
        Title,
        journal,
        year
    """
    assert extract_citation(docstring) == "Author1, Author2, Title, journal, year"


def test_extract_citation_underline():
    """Test that we can extract a citation using the underline section format."""
    docstring = """
    Top material
    ------------
    Stuff here.

    Citation
    ------
        Author1, Author2, Title, year.

    Bottom material
    ---------------
    More stuff."""
    assert extract_citation(docstring) == "Author1, Author2, Title, year."

    docstring = """Function description.

    Reference
    -----------
        Author1,
        Author2,
        Title,
        year

    Parameters
    ----------
    Stuff here.

    Returns
    -------
    More stuff."""
    assert extract_citation(docstring) == "Author1, Author2, Title, year"

    # Sections end after two blank lines.
    docstring = """Function description.

    Reference
    -----------
        Author1,
        Author2,
        Title2,
        year


    This is not part of the citation.

    Parameters
    ----------
    Stuff here.

    Returns
    -------
    More stuff."""
    assert extract_citation(docstring) == "Author1, Author2, Title2, year"


def test_extract_citation_multiple():
    """Test that we can extract multiple citations from a string."""
    # Test multi-line docstrings.
    docstring = """Top material:
    Stuff here.

    Citation: Author1, Author2, Title1, year.

    Citation: Author3, Author4, Title2, year.

    Bottom material:
    More stuff."""
    assert extract_citation(docstring) == "Author1, Author2, Title1, year.\nAuthor3, Author4, Title2, year."

    docstring = """Function description.

    Reference
    ---------
        Author1, Author2, Title1, year.

    Reference
    ---------
        Author3, Author4, Title2, year.

    Parameters
    ----------
    Stuff here.

    Returns
    -------
    More stuff."""
    assert extract_citation(docstring) == "Author1, Author2, Title1, year.\nAuthor3, Author4, Title2, year."


def test_extract_urls():
    """Test that we can extract urls from a docstring."""
    # Check an empty docstring.
    assert extract_urls("") == []
    assert extract_urls(None) == []

    # Start with single line docstrings.
    assert extract_urls("https://my_paper_url") == ["https://my_paper_url"]
    assert extract_urls("Please cite: https://my_paper_url") == ["https://my_paper_url"]
    assert extract_urls("Citation:\n    http://my_paper_url") == ["http://my_paper_url"]
    assert extract_urls("Info: Nothing to see here") == []

    # Test multi-line docstrings.
    docstring = """Top material:
    Stuff here.

    Citation:
        https://my_paper_url1
        https://my_paper_url2

    Bottom material:
    More stuff."""
    assert extract_urls(docstring) == ["https://my_paper_url1", "https://my_paper_url2"]



import fake_module

from citation_compass.citation import _get_full_name
from citation_compass import (
    cite_function,
    cite_object,
    find_in_citations,
    get_all_citations,
    get_used_citations,
    reset_used_citations,
)


@cite_function()
def example_function_1():
    """function_citation_1"""
    return 1


@cite_function()
def example_function_2():
    """function_citation_2"""
    return 2


@cite_function()
def example_function_x(x):
    """function_citation_x"""
    return x


def test_get_full_name():
    """Check that the full name is correctly generated."""
    assert _get_full_name(example_function_1) == "test_citation.example_function_1"
    assert _get_full_name(_get_full_name) == "citation_compass.citation._get_full_name"
    assert _get_full_name(fake_module.fake_uncited_function) == "fake_module.fake_uncited_function"
    assert _get_full_name(fake_module.FakeClass) == "fake_module.FakeClass"

    obj = fake_module.FakeClass()
    assert _get_full_name(obj) == "fake_module.FakeClass"
    assert _get_full_name(obj.fake_method) == "fake_module.FakeClass.fake_method"


def test_citations_all():
    """Check that all the citations are registered."""
    known_citations = [
        # The functions defined in this file.
        "test_citation.example_function_1: function_citation_1",
        "test_citation.example_function_2: function_citation_2",
        "test_citation.example_function_x: function_citation_x",
        # The items defined in fake_module.
        "fake_module: CitationCompass, 2025.",
        "fake_module.FakeClass.fake_method: A fake class method for testing.",
        "fake_module.FakeCitedClass: A 2nd fake class for testing.",
        "fake_module.InheritedFakeClass: A 3rd fake class for testing.",
    ]
    assert sorted(get_all_citations()) == sorted(known_citations)

    # Check that we have preserved the name and __doc__ string of the function
    # through the wrapping process.
    assert example_function_1.__name__ == "example_function_1"
    assert example_function_1.__doc__ == "function_citation_1"
    assert example_function_2.__name__ == "example_function_2"
    assert example_function_2.__doc__ == "function_citation_2"
    assert example_function_x.__name__ == "example_function_x"
    assert example_function_x.__doc__ == "function_citation_x"

    assert fake_module.FakeCitedClass.__name__ == "FakeCitedClass"
    assert fake_module.FakeCitedClass.__doc__ == "A 2nd fake class for testing."
    obj = fake_module.FakeCitedClass()
    assert isinstance(obj, fake_module.FakeCitedClass)

    # A citation with no docstring, but a label.
    @cite_function("function_citation_3")
    def example_function_3():
        return 3

    assert example_function_3() == 3

    # Check we have added the citation.
    known_citations.append(
        "test_citation.test_citations_all.<locals>.example_function_3: function_citation_3"
    )
    assert sorted(get_all_citations()) == sorted(known_citations)

    # We can add a citation without a label.
    @cite_function()
    def example_function_4():
        return 4

    assert example_function_4() == 4

    # Check we have added the citation.
    known_citations.append(
        "test_citation.test_citations_all.<locals>.example_function_4: No citation provided."
    )
    assert sorted(get_all_citations()) == sorted(known_citations)

    # We can add a citation without parentheses in the decorator.
    @cite_function
    def example_function_5():
        return 5

    assert example_function_5() == 5

    # Check we have added the citation.
    known_citations.append(
        "test_citation.test_citations_all.<locals>.example_function_5: No citation provided."
    )
    assert sorted(get_all_citations()) == sorted(known_citations)

    # We can add a citation that does not track used.
    @cite_function(track_used=False)
    def example_function_6():
        return 6

    assert example_function_6() == 6

    # Check we have added the citation.
    known_citations.append(
        "test_citation.test_citations_all.<locals>.example_function_6: No citation provided."
    )
    assert sorted(get_all_citations()) == sorted(known_citations)


def test_citations_used():
    """Check that the used citations are registered as they are used."""
    # Start by resetting the list of used citations, because they may
    # have been used in previous tests.
    reset_used_citations()
    used_citations = []
    assert len(get_used_citations()) == 0

    # We can use the functions as normal - add example_function_1.
    assert example_function_1() == 1
    used_citations.append("test_citation.example_function_1: function_citation_1")
    assert sorted(get_used_citations()) == sorted(used_citations)

    # We can use the functions as normal - add example_function_x.
    assert example_function_x(10) == 10
    used_citations.append("test_citation.example_function_x: function_citation_x")
    assert sorted(get_used_citations()) == sorted(used_citations)

    # Reusing a function (example_function_x) does not re-add it.
    assert example_function_x(-5) == -5
    assert example_function_x(15) == 15
    assert sorted(get_used_citations()) == sorted(used_citations)

    # Creating a new function doesn't mark it as used.
    @cite_function()
    def example_function_5():
        """Test"""
        return 5

    assert sorted(get_used_citations()) == sorted(used_citations)

    # We can use the function and it will be marked as used.
    assert example_function_5() == 5
    used_citations.append("test_citation.test_citations_used.<locals>.example_function_5: Test")
    assert sorted(get_used_citations()) == sorted(used_citations)

    # Using an uncited function doesn't add it to the list.
    _ = fake_module.fake_uncited_function()
    citations = get_used_citations()
    assert len(citations) == len(used_citations)
    for item in used_citations:
        assert item in citations

    # We can add a citation that does not track used. This function is
    # added to the "used" list immediately.
    @cite_function(track_used=False)
    def not_used():
        """Test"""
        return 6

    used_citations.append("test_citation.test_citations_used.<locals>.not_used: Test")
    assert sorted(get_used_citations()) == sorted(used_citations)
    assert not_used() == 6

    # We can manually cite an object.
    obj = fake_module.FakeClass()
    cite_object(obj)
    used_citations.append("fake_module.FakeClass: A fake class for testing.")

    # We can cite a class method.
    assert obj.fake_method() == 0
    used_citations.append("fake_module.FakeClass.fake_method: A fake class method for testing.")
    assert sorted(get_used_citations()) == sorted(used_citations)

    # The CitedClass is added to used the first time it is instantiated.
    obj2 = fake_module.FakeCitedClass()
    used_citations.append("fake_module.FakeCitedClass: A 2nd fake class for testing.")
    assert sorted(get_used_citations()) == sorted(used_citations)

    # Calling object methods does not change anything.
    assert obj2.fake_method() == 1
    assert sorted(get_used_citations()) == sorted(used_citations)

    # Instantiating the class again does not change anything.
    _ = fake_module.FakeCitedClass()
    assert sorted(get_used_citations()) == sorted(used_citations)

    # Instantiating a subclass adds that to the used list.
    _ = fake_module.InheritedFakeClass()
    used_citations.append("fake_module.InheritedFakeClass: A 3rd fake class for testing.")
    assert sorted(get_used_citations()) == sorted(used_citations)

    # We can reset the list of used citation functions.
    reset_used_citations()
    assert len(get_used_citations()) == 0


def test_find_in_citations():
    """Test the find_in_citations() function."""

    # Test we can find text by the key.
    assert len(find_in_citations("function_citation_1")) == 1
    assert len(find_in_citations("function_citation_2")) == 1
    assert len(find_in_citations("function_citation_300")) == 0
    assert len(find_in_citations("function")) >= 3
    assert len(find_in_citations("FakeCitedClass")) == 1

    # Test we can find text by the citation.
    assert len(find_in_citations("A 2nd fake class for testing.")) == 1
    assert len(find_in_citations("A fake class method for testing.")) == 1
    assert len(find_in_citations("This string does not exist.")) == 0

    # Check used citations.
    reset_used_citations()
    assert len(find_in_citations("function_citation_1", True)) == 0
    assert len(find_in_citations("FakeCitedClass", True)) == 0
    assert len(find_in_citations("A 2nd fake class for testing.", True)) == 0

    _ = example_function_1()
    assert len(find_in_citations("function_citation_1", True)) == 1

    _ = fake_module.FakeCitedClass()
    assert len(find_in_citations("FakeCitedClass", True)) == 1
